package be.iminds.ilabt.jfed.ssh_terminal_tool.ssh_key_info;

import be.iminds.ilabt.jfed.lowlevel.ssh_key_info.AbstractSshKeyInfo;
import be.iminds.ilabt.jfed.preferences.JFedPreferences;
import be.iminds.ilabt.jfed.ssh_terminal_tool.putty.PuTTYPrivateKeyFile;
import be.iminds.ilabt.jfed.util.IOUtils;
import be.iminds.ilabt.jfed.util.JFedUtils;
import be.iminds.ilabt.jfed.util.KeyUtil;
import be.iminds.ilabt.jfed.util.tmp_file_helpers.PermanentFile;
import be.iminds.ilabt.jfed.util.tmp_file_helpers.TmpContentFile;
import be.iminds.ilabt.jfed.util.tmp_file_helpers.TmpFile;
import org.slf4j.LoggerFactory;

import java.io.File;
import java.io.IOException;
import java.security.PrivateKey;
import java.security.PublicKey;
import java.security.interfaces.RSAPrivateCrtKey;
import java.security.interfaces.RSAPrivateKey;
import java.security.interfaces.RSAPublicKey;
import java.util.Date;

/**
 * PreferencesSshKeyInfo
 */
public class PreferencesSshKeyInfo extends AbstractSshKeyInfo {
    private static final org.slf4j.Logger LOG = LoggerFactory.getLogger(PreferencesSshKeyInfo.class);

    private File pubKeyFile;
    private File privKeyFile;
    private String pubKeyText;
    private String privKeyText;

    private PublicKey publicKey;
    private PrivateKey privateKey;

    private final boolean complete;

    public PreferencesSshKeyInfo(JFedPreferences jFedPreferences,
                                 JFedPreferences.PreferenceKey overridePref,
                                 JFedPreferences.PreferenceKey pubKeyFilePref,
                                 JFedPreferences.PreferenceKey privKeyFilePref,
                                 JFedPreferences.PreferenceKey pubKeyTextPref,
                                 JFedPreferences.PreferenceKey privKeyTextPref) {
        boolean overrideVal = jFedPreferences.getBoolean(overridePref);
        assert overrideVal;

        pubKeyFile = jFedPreferences.getFile(pubKeyFilePref);
        privKeyFile = jFedPreferences.getFile(privKeyFilePref);

        pubKeyText = jFedPreferences.getString(pubKeyTextPref);
        privKeyText = jFedPreferences.getString(privKeyTextPref);

      //  assert privKeyFile != null || privKeyText != null;
        if (privKeyFile == null && privKeyText == null) {
            LOG.debug("No private key");
            complete = false;
            return;
        }

        if (privKeyText == null) {
            try {
                privKeyText = IOUtils.fileToString(privKeyFile);
            } catch (IOException e) {
                LOG.debug("Could not open private key file from preferences: "+privKeyFile.getPath()+" error: "+e.getMessage(), e);
                complete = false;
                return;
            }
        }

        if (KeyUtil.hasEncryptedRsaPrivateKey(privKeyText)) {
            //TODO handle encrypted RSA private key
        } else {
            //parse privKeyText into privateKey
            try {
                privateKey = KeyUtil.pemToAnyPrivateKey(privKeyText, null);
            } catch (KeyUtil.PEMDecodingException e) {
                LOG.debug("Could not parse private key from preferences (privKeyFile="+privKeyFile+" privKeyText head="+(privKeyText == null ? null : privKeyText.substring(0, 50))+"). error: " + e.getMessage(), e);
                complete = false;
                return;
            }
        }

        //check if public key can be derived from private key
        if (privateKey != null && privateKey instanceof RSAPrivateKey) {
            RSAPrivateKey rsaPrivateKey = (RSAPrivateKey) privateKey;
            publicKey = KeyUtil.rsaPrivateKeyToRsaPublicKey(rsaPrivateKey);
        }

        if (publicKey == null) {
            if (pubKeyFile == null && pubKeyText == null) {
                LOG.debug("No public key");
                complete = false;
                return;
            }

            //if not, load public key file
            if (pubKeyText == null) {
                try {
                    pubKeyText = IOUtils.fileToString(pubKeyFile);
                } catch (IOException e) {
                    LOG.debug("Could not open public key file from preferences: "+pubKeyFile.getPath()+" error: "+e.getMessage(), e);
                    complete = false;
                    return;
                }
            }

            //and process publicKey
            try {
                publicKey = KeyUtil.openSshAuthorizedKeysFormatRsaPublicKey(pubKeyText);
            } catch (Exception e) {
                LOG.debug("Could not parse public key from preferences error: " + e.getMessage(), e);
                complete = false;
                return;
            }

            //TODO: check if matches private key
            if ((privateKey instanceof RSAPrivateCrtKey) && (publicKey instanceof RSAPublicKey)) {
                RSAPrivateCrtKey rsaPrivateKey = (RSAPrivateCrtKey) privateKey;
                if (!rsaPrivateKey.getPublicExponent().equals(((RSAPublicKey) publicKey).getPublicExponent())) {
                    LOG.debug("Public and private key do not match!");
                    complete = false;
                    return;
                }
            }
        } else {
            //ignore public key in preferences and use the one matching the private key
            LOG.debug("Ignoring public key as private key contains public key");
            pubKeyFile = null;
            pubKeyText = null;
//            pubKeyText = KeyUtil.publicKeyToOpenSshAuthorizedKeysFormat(publicKey); //no need for this
        }

        //this class is only complete when both keys have been read
//        assert privateKey != null;
        assert publicKey != null;

        //all complete
        complete = true;
    }


    @Override
    public TmpFile getPrivateKeyFile() {
        if (privKeyFile != null) {
            PermanentFile res = new PermanentFile(privKeyFile);
            ensureMatchingPubFile(res);
            return res;
        }

        File tmpDir = new File(JFedUtils.getUserDataDirectoryFile(), "tmp");
        if (!tmpDir.exists()) tmpDir.mkdir();
        long nr = new Date().getTime();
        TmpContentFile tmpPrivateKeyFile;

        if (privKeyText != null) {
            tmpPrivateKeyFile = new TmpContentFile("sskKeyPrefE", "pem", privKeyText, true/*assureUserOnlyPermissions*/);
        }
        else {
            //this does not normally occur
            tmpPrivateKeyFile = new TmpContentFile("sskKeyPrefE", "pem", new String(KeyUtil.privateKeyToAnyPem(privateKey)), true/*assureUserOnlyPermissions*/);
        }
        ensureMatchingPubFile(tmpPrivateKeyFile);
        return tmpPrivateKeyFile;
    }


    @Override
    public TmpFile getPuttyKeyFile() {
        if (publicKey == null || privateKey == null) return null;

        return getUnlockedPuttyKeyFile();
    }

    @Override
    public TmpFile getUnencryptedPrivateKeyFile() {
        if (privKeyText != null && !KeyUtil.hasEncryptedRsaPrivateKey(privKeyText))
            return getPrivateKeyFile();

        if (privateKey == null)
            return null;

        String pemKey = privKeyText;
        if (privKeyText == null)
            privKeyText = new String(KeyUtil.privateKeyToAnyPem(privateKey));
        TmpContentFile unlockedPrivateKeyFile = new TmpContentFile("sskKeyPrefUE", ".pem", pemKey, true/*assureUserOnlyPermissions*/);

        ensureMatchingPubFile(unlockedPrivateKeyFile);
        return unlockedPrivateKeyFile;
    }

    @Override
    public TmpFile getUnlockedPuttyKeyFile() {
        if (publicKey == null || privateKey == null) return null;

        String ppkContent = PuTTYPrivateKeyFile.createContent("jFed-tmp-ppk", (RSAPublicKey) publicKey, (RSAPrivateCrtKey) privateKey, null);
        TmpFile unlockedPuttyFile = new TmpContentFile("sskKeyPrefUE", ".ppk", ppkContent);

        return unlockedPuttyFile;
    }

    @Override
    public boolean hasUnencryptedPrivateKey() {
        return privateKey != null;
    }

//    @Override
//    public void release() {
//        for (File matchingPubKeyfile : matchingPubKeyfiles) {
//            matchingPubKeyfile.delete();
//        }
//        matchingPubKeyfiles.clear();
//    }

    @Override
    public PublicKey getPublicKey() {
        return publicKey;
    }

    @Override
    public PrivateKey getPrivateKey() {
        return privateKey;
    }

    public boolean isComplete() {
        return complete;
    }
}
