package be.iminds.ilabt.jfed.connectivity_tester;

import be.iminds.ilabt.jfed.lowlevel.GeniUserProvider;
import be.iminds.ilabt.jfed.lowlevel.connection.JFedConnection;
import be.iminds.ilabt.jfed.lowlevel.ssh_key_info.SshKeyInfo;
import be.iminds.ilabt.jfed.preferences.JFedGuiPreferences;
import be.iminds.ilabt.jfed.ssh_terminal_tool.ssh_key_info.UserSshKeyInfo;
import be.iminds.ilabt.jfed.util.*;
import com.jcraft.jsch.*;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.net.ssl.*;
import java.io.*;
import java.net.InetSocketAddress;
import java.net.Socket;
import java.net.SocketException;
import java.security.KeyManagementException;
import java.security.KeyStore;
import java.security.KeyStoreException;
import java.security.NoSuchAlgorithmException;

/**
 * User: twalcari
 * Date: 3/5/14
 * Time: 4:09 PM
 */
public class SSHLoginTest implements ConnectivityTest {
    private static final Logger LOG = LoggerFactory.getLogger(SSHLoginTest.class);
    private final String hostname;
    private final int port;
    private final String username;
    private final JFedConnection.ProxyInfo proxyInfo;

    //FIXME: how can we inject this value? We are loading this class dynamically...
    private final GeniUserProvider geniUserProvider;
    private final JFedGuiPreferences jFedPreferences;

    /**
     * @param hostname
     * @param port
     * @param username
     * @param proxyInfo set to proxyInfo when needed. null when no proxy. Hint: guiModel.getSfaConnectionPool().getProxyInfo();
     * @param geniUserProvider
     */
    public SSHLoginTest(String hostname, int port, String username, JFedConnection.ProxyInfo proxyInfo, GeniUserProvider geniUserProvider, JFedGuiPreferences jFedPreferences) {
        this.hostname = hostname;
        this.port = port;
        this.username = username;
        this.proxyInfo = proxyInfo;
        this.geniUserProvider = geniUserProvider;
        this.jFedPreferences = jFedPreferences;
    }

    @Override
    public String getName() {
        return "SSH Login " + username + "@" + hostname + ":" + port + (proxyInfo == null ? ""
                : " with proxy " + proxyInfo);
    }

    @Override
    public boolean isLoginRequired() {
        return true;
    }

    @Override
    public ConnectivityTestResult call() {
        SshKeyInfo sshKeyInfo = jFedPreferences.getOverriddenSshKeyInfo();
        if (sshKeyInfo != null && !sshKeyInfo.hasUnencryptedPrivateKey()) {
            LOG.info("Private key in preferences is encrypted. Login cannot be tested."+
                    " sshKeyInfo.getPrivateKey()="+sshKeyInfo.getPrivateKey()+
                    " sshKeyInfo.hasUnencryptedPrivateKey()="+sshKeyInfo.hasUnencryptedPrivateKey());
            return new ConnectivityTestResult(Status.SKIPPED, "Private key in preferences is encrypted. Login cannot be tested.");
        }
        if (sshKeyInfo == null) {
            LOG.info("The user's SSH key will be used.");
            sshKeyInfo = new UserSshKeyInfo(geniUserProvider.getLoggedInGeniUser());
        }
        if (sshKeyInfo == null || !sshKeyInfo.hasUnencryptedPrivateKey() || sshKeyInfo.getPrivateKey() == null) {
            LOG.info("Private key in preferences is encrypted. Login cannot be tested."+"sshKeyInfo == "+
                (sshKeyInfo == null ? "null" :
                    ("notnull"+
                     " sshKeyInfo.hasUnencryptedPrivateKey()="+sshKeyInfo.hasUnencryptedPrivateKey()+
                     " sshKeyInfo.getPrivateKey()="+(sshKeyInfo.getPrivateKey() == null ? "null" : "notnull")
                    )
                )
            );
            return new ConnectivityTestResult(Status.SKIPPED, "Private key in preferences is encrypted ("+
                    (sshKeyInfo == null ? "null" :
                            (
                                sshKeyInfo.hasUnencryptedPrivateKey() ? ("false, but privKey="+(sshKeyInfo.getPrivateKey() == null ? "null" : "notnull")) : "true"
                            )
                    )
                    +"). Login cannot be tested.");
        }

        JSch jSch = new JSch();
        try {
            jSch.addIdentity("TEST",
                    new String(KeyUtil.privateKeyToAnyPem(sshKeyInfo.getPrivateKey())).getBytes("US-ASCII"), null, null);


        } catch (UnsupportedEncodingException | JSchException e) {
            LOG.error("Error while registring private key", e);
            return new ConnectivityTestResult(Status.FAILED, "Received an error while registering private key", e);
        }

        Session session = null;
        try {
            session = jSch.getSession(username, hostname, port);

            //configure jSch to accept any host key
            java.util.Properties config = new java.util.Properties();
            config.put("StrictHostKeyChecking", "no");
            session.setConfig(config);

            if (proxyInfo != null) {
                if (proxyInfo instanceof JFedConnection.SshProxyInfo) {
                    LOG.info("Using SSH proxy for SSH connection: " + proxyInfo);
                    session.setProxy(new SSHProxy((JFedConnection.SshProxyInfo) proxyInfo));
                } else {
                    session.setProxy(new ProxySOCKS4(proxyInfo.getHostname(), proxyInfo.getPort()));
                    LOG.info("Using ProxySOCKS4 for SSH connection: " + proxyInfo);
                }
            }

            session.connect();

            if (session.isConnected()) {
                LOG.debug("Successfully authenticated {}@{}:{}", username, hostname, port);

                Channel channel = session.openChannel("exec");

                ((ChannelExec) channel).setCommand("uname -a");

                channel.setInputStream(null);
                ByteArrayOutputStream errStream = new ByteArrayOutputStream();
                ((ChannelExec) channel).setErrStream(errStream);
                ByteArrayOutputStream outStream = new ByteArrayOutputStream();
                channel.setOutputStream(outStream);

                channel.connect();

                while (!channel.isClosed()) {
                    LOG.debug("Waiting for test-commands to finish..");
                    try {
                        Thread.sleep(100);
                    } catch (InterruptedException ignored) {
                    }
                }
                LOG.debug("Test-channel exited with status " + channel.getExitStatus());

                LOG.info("Output of test-commands: " + outStream.toString());
                if (errStream.size() > 0)
                    LOG.warn("Error-output of test-commands: " + errStream.toString());

                return new ConnectivityTestResult(Status.SUCCEEDED,
                        "Successfully authenticated " + username + "@" + hostname + ":" + port);
            } else {
                LOG.debug("Could not authenticate to {}@{}:{}", username, hostname, port);
                return new ConnectivityTestResult(Status.FAILED,
                        "Could not authenticate to " + username + "@" + hostname + ":" + port);
            }


        } catch (JSchException e) {
            LOG.error("Got error while testing connectivity to node", e);
            return new ConnectivityTestResult(Status.FAILED, "Received an error while connecting", e);
        } finally {
            if (session != null)
                session.disconnect();
        }
    }


    private static class SSHProxy implements Proxy {
        private final JFedConnection.SshProxyInfo proxyInfo;
        SSLSocketFactory factory;
        private Socket underlying;

        public SSHProxy(JFedConnection.SshProxyInfo proxyInfo) {
            this.proxyInfo = proxyInfo;
            KeyStore systemTrustStore = JFedTrustStore.getSystemTrustStore();

            SSLContext context = null;
            try {
                context = SSLContext.getInstance("TLS");
                TrustManagerFactory tmf =
                        TrustManagerFactory.getInstance(TrustManagerFactory.getDefaultAlgorithm());
                tmf.init(systemTrustStore);
                X509TrustManager defaultTrustManager = (X509TrustManager) tmf.getTrustManagers()[0];
                SSLCertificateDownloader.SavingTrustManager tm =
                        new SSLCertificateDownloader.SavingTrustManager(defaultTrustManager);
                context.init(null, new TrustManager[]{tm}, null);
                factory = context.getSocketFactory();
            } catch (NoSuchAlgorithmException e) {
                e.printStackTrace();
            } catch (KeyStoreException e) {
                e.printStackTrace();
            } catch (KeyManagementException e) {
                e.printStackTrace();
            }
        }


        @Override
        public void connect(SocketFactory socketFactory, String host, int port, int timeout) throws Exception {
            JFedConnection.SshProxyInfo sshProxyInfo = proxyInfo;
            SshServerProxyHelper.SshProxyInfo sshServerProxyInfo =
                    new SshServerProxyHelper.SshProxyInfo(
                            new InetSocketAddress(sshProxyInfo.getHostname(), sshProxyInfo.getPort()),
                            sshProxyInfo.getUsername(),
                            sshProxyInfo.getHostKey(),
                            new String(KeyUtil.privateKeyToAnyPem(sshProxyInfo.getSshKeyInfo().getPrivateKey())));

            GanymedSshSocketImplFactory sshHSocketImplFactory =
                    GanymedSshSocketImplFactory.getFactory(sshServerProxyInfo);

            underlying = new SSHSocket(sshHSocketImplFactory);
            underlying.connect(new InetSocketAddress(host, port), 10000);
        }

        @Override
        public InputStream getInputStream() {
            try {
                return underlying.getInputStream();
            } catch (IOException e) {
                e.printStackTrace();
                return null;
            }
        }

        @Override
        public OutputStream getOutputStream() {
            try {
                return underlying.getOutputStream();
            } catch (IOException e) {
                e.printStackTrace();
                return null;
            }
        }

        @Override
        public Socket getSocket() {
            return underlying;
        }

        @Override
        public void close() {

        }

        private static class SSHSocket extends Socket {
            public SSHSocket(GanymedSshSocketImplFactory ganymedSshSocketImplFactory) throws SocketException {
                super(ganymedSshSocketImplFactory.createSocketImpl());
            }
        }


    }
}
