package be.iminds.ilabt.jfed.experimenter_gui.controller.view;

import be.iminds.ilabt.jfed.experimenter_gui.controller.controller.ControllerController;
import be.iminds.ilabt.jfed.experimenter_gui.controller.model.ControllerModel;
import be.iminds.ilabt.jfed.experimenter_gui.controller.view.output.ControllerOutput;
import be.iminds.ilabt.jfed.experimenter_gui.editor.ModelRspecEditor;
import be.iminds.ilabt.jfed.experimenter_gui.slice.DetailsTabHost;
import be.iminds.ilabt.jfed.experiment.Experiment;
import be.iminds.ilabt.jfed.experimenter_gui.slice.NodePropertiesDialogFactory;
import be.iminds.ilabt.jfed.experimenter_gui.util.ImageUtil;
import be.iminds.ilabt.jfed.experimenter_gui.util.ProxyServiceUtil;
import be.iminds.ilabt.jfed.highlevel.model.Slice;
import be.iminds.ilabt.jfed.lowlevel.GeniUserProvider;
import be.iminds.ilabt.jfed.preferences.JFedGuiPreferences;
import be.iminds.ilabt.jfed.rspec.model.ModelRspecType;
import be.iminds.ilabt.jfed.rspec.model.controller.ExperimentBarrierSegment;
import be.iminds.ilabt.jfed.rspec.model.controller.ExperimentCommand;
import be.iminds.ilabt.jfed.rspec.model.javafx_impl.FXModelRspec;
import be.iminds.ilabt.jfed.rspec.model.javafx_impl.FXRspecNode;
import be.iminds.ilabt.jfed.rspec.rspec_source.ManifestRspecSource;
import be.iminds.ilabt.jfed.util.ProxyPreferencesManager;
import javafx.beans.value.ObservableBooleanValue;
import javafx.geometry.Orientation;
import javafx.scene.control.SplitPane;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * @author Cedric
 */
public class ControllerUI extends SplitPane implements IControllerComponent {
    private static final Logger LOG = LoggerFactory.getLogger(ControllerUI.class);


    private final FXModelRspec modelRspec;
    private final ControllerModel model;
    private final ControllerController controller;
    private final ControllerTimeline experimentController;
    private final ControllerTreeView experimentList;
    private ControllerOutput experimentOutput;
    private Slice slice;

    /**
     * Constructor which initiates ("offline") editor-view of controller.
     *  @param modelRspecEditor
     * @param nodePropertiesDialog
     * @param imageUtil
     * @param geniUserProvider
     * @param proxyPreferencesManager
     * @param proxyServiceUtil
     * @param controllerNodeFactory
     */
    public ControllerUI(ModelRspecEditor modelRspecEditor, NodePropertiesDialogFactory nodePropertiesDialog,
                        ImageUtil imageUtil, GeniUserProvider geniUserProvider,
                        ProxyPreferencesManager proxyPreferencesManager, ProxyServiceUtil proxyServiceUtil,
                        ControllerNodeFactory controllerNodeFactory, JFedGuiPreferences jFedPreferences) {
        //this.modelRspecEditor = modelRspecEditor;
        this.modelRspec = modelRspecEditor.getModelRspec();
        controller = new ControllerController(jFedPreferences);
        model =
                new ControllerModel(modelRspecEditor.getModelRspec(), ControllerModel.CommandEditorMode.OFFLINE_EDITOR,
                        geniUserProvider, proxyPreferencesManager, proxyServiceUtil, controllerNodeFactory);
        controller.setModel(model);
        controller.setView(this);

        experimentController = new ControllerTimeline(model, controller);
        experimentList = new ControllerTreeView(model, controller, nodePropertiesDialog, imageUtil);


        this.setDividerPosition(0, 0.2);
        this.setOrientation(Orientation.HORIZONTAL);
        this.getItems().addAll(experimentList, experimentController);

        SplitPane.setResizableWithParent(experimentList, false);

        initializeTimelineGui();

    }

    /**
     * Constructor for use as an "online" experiment recorder
     */
    public ControllerUI(Experiment experiment, ManifestRspecSource manifestRspecSource, DetailsTabHost detailsTabHost,
                        NodePropertiesDialogFactory nodePropertiesDialog, ImageUtil imageUtil, GeniUserProvider geniUserProvider,
                        ProxyPreferencesManager proxyPreferencesManager, ProxyServiceUtil proxyServiceUtil,
                        ControllerNodeFactory controllerNodeFactory, JFedGuiPreferences jFedPreferences) {
        this.modelRspec = (FXModelRspec) manifestRspecSource.getModelRspec(ModelRspecType.FX);
        controller = new ControllerController(jFedPreferences);
        model =
                new ControllerModel(modelRspec/*requestModelRspec*/, ControllerModel.CommandEditorMode.ONLINE_EDITOR,
                        geniUserProvider, proxyPreferencesManager, proxyServiceUtil, controllerNodeFactory);
        controller.setModel(model);
        controller.setView(this);

        experimentController = new ControllerTimeline(model, controller);
        experimentList = new ControllerTreeView(model, controller, nodePropertiesDialog, imageUtil);

        experimentOutput = new ControllerOutput(detailsTabHost, controller, imageUtil);


        getItems().addAll(experimentList, experimentController);
        setOrientation(Orientation.HORIZONTAL);
        setDividerPosition(0, 0.2);


        SplitPane.setResizableWithParent(experimentList, false);

        initializeTimelineGui();

        model.setExperiment(experiment);
    }

    private void initializeTimelineGui() {
        if (modelRspec.getBarrierSegments().isEmpty()) {
            controller.addBarrierSegment();
        } else {
            for (ExperimentBarrierSegment b : modelRspec.getBarrierSegments()) {
                update(UpdateEvent.ADD_BARRIER, b);
            }
        }

        for (FXRspecNode n : model.getNodes()) {
            update(UpdateEvent.ADD_NODE, n);
            for (ExperimentCommand c : n.getCommands()) {
                LOG.debug("initializeTimelineGui() for command for node with id " + c.getNodeUniqueID() + " on node with id " + n.getUniqueId());
                assert n.getUniqueId().equals(c.getNodeUniqueID()) :
                        "node id\"" + n.getUniqueId() + "\" does not equal command node id \"" + c.getNodeUniqueID() + "\"";
                update(UpdateEvent.ADD_COMMAND, c);
            }
        }
    }

    @Override
    public void update(UpdateEvent event, Object o) {
        experimentList.update(event, o);
        experimentController.update(event, o);

        //experimentOutput is not used in the editor mode
        if (model.getEditorMode().equals(ControllerModel.CommandEditorMode.ONLINE_EDITOR)) {
            experimentOutput.update(event, o);
        }
    }

    public FXModelRspec getModelRspec() {
        return modelRspec;
    }

    public ControllerController getController() {
        return controller;
    }

    public ControllerModel getModel() {
        return model;
    }

    public ObservableBooleanValue onlineProperty() {
        return model.allConnectableProperty();
    }


}
