package be.iminds.ilabt.jfed.experimenter_gui.editor.properties;

import be.iminds.ilabt.jfed.experimenter_gui.JFedConfiguration;
import be.iminds.ilabt.jfed.experimenter_gui.editor.properties.cmi.CMIManager;
import be.iminds.ilabt.jfed.experimenter_gui.editor.properties.cmi.ComponentManagerInfo;
import be.iminds.ilabt.jfed.experimenter_gui.editor.properties.map.MapDialogFactory;
import be.iminds.ilabt.jfed.experimenter_gui.util.BonfireLicenseUtil;
import be.iminds.ilabt.jfed.experimenter_gui.util.StringConverters;
import be.iminds.ilabt.jfed.lowlevel.JFedConnectivityException;
import be.iminds.ilabt.jfed.rspec.model.HardwareType;
import be.iminds.ilabt.jfed.rspec.model.RspecNode;
import be.iminds.ilabt.jfed.rspec.model.javafx_impl.FXModelRspec;
import be.iminds.ilabt.jfed.rspec.model.javafx_impl.FXRspecFactory;
import be.iminds.ilabt.jfed.rspec.model.javafx_impl.FXRspecNode;
import be.iminds.ilabt.jfed.ui.javafx.FXMLUtil;
import be.iminds.ilabt.jfed.ui.javafx.dialogs.JFDialogs;
import com.google.common.collect.ImmutableSet;
import javafx.beans.InvalidationListener;
import javafx.beans.Observable;
import javafx.beans.property.ListProperty;
import javafx.beans.property.ObjectProperty;
import javafx.beans.property.ReadOnlyBooleanProperty;
import javafx.beans.property.SimpleListProperty;
import javafx.beans.property.SimpleObjectProperty;
import javafx.beans.value.ChangeListener;
import javafx.beans.value.ObservableValue;
import javafx.event.ActionEvent;
import javafx.fxml.FXML;
import javafx.scene.control.*;
import javafx.scene.layout.StackPane;
import javafx.scene.layout.VBox;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.Comparator;
import java.util.List;
import java.util.Optional;
import java.util.stream.Collectors;

/**
 * User: twalcari
 * Date: 8/29/2014
 * Time: 11:26
 */
public class ComponentManagerPropertiesPane extends StackPane {
    private final static Logger LOG = LoggerFactory.getLogger(ComponentManagerPropertiesPane.class);
    /**
     * Blacklists as mentioned in FEDIBBTDEV-2100
     */
    private static final ImmutableSet<String> HARDWARE_TYPE_NAMES_SUBSTRING_BLACKLIST =
            ImmutableSet.of("80211", "flex900", "xcvr2450", "delay", "interconnect", "ipv4", "lan", "vm", "switch");
    private static final ImmutableSet<String> HARDWARE_TYPE_NAMES_BLACKLIST =
            ImmutableSet.of("pc");

    private static final Comparator<HardwareType> HARDWARE_TYPE_COMPARATOR =
            Comparator.comparing(HardwareType::getName, String::compareToIgnoreCase);

    private final CMIManager cmiManager;
    private final JFedConfiguration config;
    private final MapDialogFactory mapDialogFactory;
    private final ComponentManagerInfo componentManagerInfo;
    private final JFedConfiguration.TestbedDescription testbedDescription;
    private final BonfireLicenseUtil bonfireLicenseUtil;

    @FXML
    private ComboBox<FXRspecNode> componentIdComboBox;
    @FXML
    private ComboBox<HardwareType> hardwareTypesComboBox;
    @FXML
    private ToggleGroup selectComponentIdToggleGroup;
    @FXML
    private RadioButton componentIdNodeRadioButton, anyNodeRadioButton, hardwareTypeNodeRadioButton;
    @FXML
    private ProgressBar advertisementFetchProgressBar;
    @FXML
    private VBox advertisementFetchOverlay;
    @FXML
    private Button nodesVizButton;

    @FXML
    private CheckBox onlyAvailableNodes;


    private final ObjectProperty<ComponentManagerInfo.Resources.RefreshAdvertisementTask> activeRefreshTask =
            new SimpleObjectProperty<>();
    private final ObjectProperty<ComponentManagerInfo.Resources> activeResources = new SimpleObjectProperty<>();


    private final ListProperty<FXRspecNode> nodesList = new SimpleListProperty<>();


    private final FXModelRspec userSpecifiedNodesModel = FXRspecFactory.getInstance().createModelRspec("advertisement");
    private final FXRspecNode userSpecifiedAdvNode;

    public ComponentManagerPropertiesPane(final ComponentManagerInfo componentManagerInfo,
                                          FXRspecNode node, boolean availableNodesOnly,
                                          CMIManager cmiManager, JFedConfiguration config, MapDialogFactory mapDialogFactory,
                                          BonfireLicenseUtil bonfireLicenseUtil) {
        this.componentManagerInfo = componentManagerInfo;
        this.cmiManager = cmiManager;
        this.config = config;
        this.bonfireLicenseUtil = bonfireLicenseUtil;
        this.testbedDescription = config.getTestbedDescription(componentManagerInfo.getAuthority().getUrn());
        this.mapDialogFactory = mapDialogFactory;

        //initialize FXML
        FXMLUtil.injectFXML(this);

        onlyAvailableNodes.setSelected(availableNodesOnly);
        //the order of bindings is important!


        //bind componentId-combobox to advertisement
        nodesList.addListener(new InvalidationListener() {
            @Override
            public void invalidated(Observable observable) {
                //backup selected node
                FXRspecNode selectedValue = componentIdComboBox.getValue();

                //update list
                componentIdComboBox.setItems(nodesList.get());
                if (userSpecifiedAdvNode != null)
                    componentIdComboBox.getItems().add(userSpecifiedAdvNode);

                //try to restore the selected value
                if (selectedValue != null) {
                    assert (selectedValue.getComponentId() != null);
                    Optional<FXRspecNode> newSelectedValue = componentIdComboBox.getItems().stream()
                            .filter(newAdvNode -> selectedValue.getComponentId().equals(newAdvNode.getComponentId()))
                            .findFirst();

                    componentIdComboBox.setValue(newSelectedValue.orElse(null));
                }
            }
        });

        //bind hardwaretypes-combobox to advertisement
        nodesList.addListener(new InvalidationListener() {
            @Override
            public void invalidated(Observable observable) {
                List<HardwareType> hardwareTypes =
                        nodesList.get().stream().flatMap(node -> node.getHardwareTypes().stream())
                                .filter(hardwareType ->
                                        !HARDWARE_TYPE_NAMES_SUBSTRING_BLACKLIST.stream()
                                                .anyMatch(blacklistItem ->
                                                        hardwareType.getName().toLowerCase()
                                                                .contains(blacklistItem))
                                                &&
                                                !HARDWARE_TYPE_NAMES_BLACKLIST.contains(hardwareType.getName().toLowerCase())
                                )
                                .distinct()
                                .sorted(HARDWARE_TYPE_COMPARATOR)
                                .collect(Collectors.toList());

                if (!hardwareTypes.contains(hardwareTypesComboBox.getSelectionModel().getSelectedItem()))
                    hardwareTypesComboBox.getSelectionModel().clearSelection();


                hardwareTypesComboBox.getItems().setAll(hardwareTypes);
            }
        });

        //1. bind refresh overlay to activeRefreshTask
        advertisementFetchOverlay.visibleProperty().unbind();
        advertisementFetchOverlay.visibleProperty().bind(activeRefreshTask.isNotNull());

        activeRefreshTask.addListener((observable, oldValue, newValue) -> {
            if (oldValue != null) {
                advertisementFetchProgressBar.progressProperty().unbind();
            }

            if (newValue != null) {
                advertisementFetchProgressBar.progressProperty().bind(newValue.progressProperty());
            }
        });

        //2. bind activeRefreshTask to activeResources
        activeResources.addListener(new ChangeListener<ComponentManagerInfo.Resources>() {
            @Override
            public void changed(ObservableValue<? extends ComponentManagerInfo.Resources> observable,
                                ComponentManagerInfo.Resources oldValue, ComponentManagerInfo.Resources newValue) {
                activeRefreshTask.unbind();
                activeRefreshTask.bind(newValue.currentRefreshAdvertisementTaskProperty());

                nodesList.set(cmiManager.filterComponentIds(componentManagerInfo, activeResources.get().getAdvertisementComponentIds()));
            }
        });

        //3. bind activeResources
        activeResources.set(availableNodesOnly ? componentManagerInfo.getAvailableResources()
                : componentManagerInfo.getAllResources());

        //initialize view
        componentIdComboBox.setConverter(StringConverters.ADV_RSPECNODE_CONVERTER);
        hardwareTypesComboBox.setConverter(StringConverters.HARDWARE_TYPE_CONVERTER);

        //initialize bindings
        componentIdComboBox.disableProperty().bind(componentIdNodeRadioButton.selectedProperty().not());
        onlyAvailableNodes.disableProperty().bind(anyNodeRadioButton.selectedProperty());

        hardwareTypesComboBox.disableProperty().bind(hardwareTypeNodeRadioButton.selectedProperty().not());

        onlyAvailableNodes.selectedProperty().addListener(observable -> {
            if (onlyAvailableNodes.isSelected()) {
                this.activeResources.setValue(componentManagerInfo.getAvailableResources());
            } else {
                this.activeResources.setValue(componentManagerInfo.getAllResources());
            }

            if (!this.activeResources.get().isComponentIdsAvailable())
                refreshAdvertisement();

        });

        //initialize pane with values from node
        if (node.getComponentId() != null) {
            Optional<FXRspecNode> nodeInList = componentIdComboBox.getItems().stream()
                    .filter(advNode -> node.getComponentId().equals(advNode.getComponentId()))
                    .findFirst();

            if (nodeInList.isPresent()) {
                componentIdComboBox.getSelectionModel().select(nodeInList.get());
                componentIdNodeRadioButton.setSelected(true);

                userSpecifiedAdvNode = null;
            } else {
                //make a copy of the node, so we can store it for the list.
                userSpecifiedAdvNode =
                        FXRspecFactory.getInstance().copyNode(userSpecifiedNodesModel, node, RspecNode.InterfaceCopyMethod.NO_COPY);
                userSpecifiedNodesModel.addNode(userSpecifiedAdvNode);

                componentIdComboBox.getItems().add(0, userSpecifiedAdvNode);

                //noinspection ConstantConditions
                assert node.getComponentId().equals(userSpecifiedAdvNode.getComponentId());

                componentIdComboBox.getSelectionModel().select(userSpecifiedAdvNode);
                componentIdNodeRadioButton.setSelected(true);
            }
        } else if (!node.getHardwareTypes().isEmpty()) {
            userSpecifiedAdvNode = null;

            if (node.getHardwareTypes().size() > 1) {
                LOG.warn("A node can only have one hardware type in request, but we found {}. Taking first one, and ignoring the rest!", node.getHardwareTypes().size());
            }

            HardwareType userSpecifiedHardwareType = node.getHardwareTypes().get(0);

            if (!hardwareTypesComboBox.getItems().contains(userSpecifiedHardwareType)) {
                hardwareTypesComboBox.getItems().add(0, userSpecifiedHardwareType);
            }
            hardwareTypesComboBox.getSelectionModel().select(userSpecifiedHardwareType);
            hardwareTypeNodeRadioButton.setSelected(true);
        } else {
            userSpecifiedAdvNode = null;
        }

        //add this listener AFTER  the possible creation of the userSpecifiedAdvNode. This to prevent an unnecessary
        //advertisement call
        componentIdNodeRadioButton.selectedProperty().addListener(observable -> {
            if (componentIdNodeRadioButton.isSelected() &&
                    !activeResources.get().isComponentIdsAvailable()) {
                refreshAdvertisement();
            }
        });

        hardwareTypeNodeRadioButton.selectedProperty().addListener(observable -> {
            if (hardwareTypeNodeRadioButton.isSelected() &&
                    !activeResources.get().isComponentIdsAvailable()) {
                refreshAdvertisement();
            }
        });

        //but make sure that a refresh is done if a fixed node assignment is mandatory!
        if (componentManagerInfo.isRequiresFixedNodeAssignment()) {
            anyNodeRadioButton.setDisable(true);
            componentIdNodeRadioButton.setSelected(true);
        }


        //check if viz-button must be enabled for this testbed
        if (testbedDescription != null && testbedDescription.getNodesVisualization() != null) {
            nodesVizButton.disableProperty().bind(componentIdComboBox.disableProperty());

            if (testbedDescription.getNodesVisualization() == JFedConfiguration.NodesVisualization.MAP) {
                nodesVizButton.setOnAction(this::showNodesMap);

                //only show maps-button if running on a 64-bit JVM (see FEDIBBTDEV-1344)
                String arch = System.getProperty("os.arch");

                if (arch.equals("x86")) {
                    nodesVizButton.setVisible(false);
                }
            } else if (testbedDescription.getNodesVisualization() == JFedConfiguration.NodesVisualization.IMAGE) {
                nodesVizButton.setOnAction(this::showNodesImageMap);
            }
        }

    }


    private void showNodesImageMap(ActionEvent actionEvent) {
        assert (testbedDescription.getImageMap() != null);
        FXRspecNode selectedNode =
                mapDialogFactory.showNodeImageMapDialog(this.getScene().getWindow(), testbedDescription.getImageMap(),
                        activeResources.get().getAdvertisementComponentIds());

        if (selectedNode != null) {
            componentIdComboBox.setValue(selectedNode);
        }
    }

    private void showNodesMap(ActionEvent actionEvent) {
        FXRspecNode selectedNode =
                mapDialogFactory.showNodeMapDialog(this.getScene().getWindow(), activeResources.get().getAdvertisementComponentIds());

        if (selectedNode != null) {
            componentIdComboBox.setValue(selectedNode);
        }
    }


    private static boolean isConnectivityException(Throwable t) {

        while (t != null) {
            if (t.getCause() instanceof JFedConnectivityException)
                return true;
            else
                t = t.getCause();
        }
        return false;
    }


    private void refreshAdvertisement() {
        final ComponentManagerInfo.Resources.RefreshAdvertisementTask updateAdvTask =
                activeResources.get().refreshAdvertisement();

        updateAdvTask.setOnFailed(workerStateEvent -> {
            //only show messages if pane is still visible
            if (ComponentManagerPropertiesPane.this.getParent() != null) {
                if (isConnectivityException(updateAdvTask.getException())) {
                    JFDialogs.create()
                            .owner(ComponentManagerPropertiesPane.this)
                            .message(String.format("A connectivity exception occurred while fetching the advertisement RSpec from %s.\n\n" +
                                            "Verify your internet connection, and consider enabling a proxy in the jFed Preferences.",
                                    componentManagerInfo.getAuthority().getName()))
                            .masthead("Fetching node advertisement failed")
                            .title("Fetching node advertisement failed")
                            .showException(updateAdvTask.getException());


                } else if(BonfireLicenseUtil.isBonfireLicenseException(updateAdvTask.getException())) {
                    bonfireLicenseUtil.showBonfireLicenseAlert(updateAdvTask.getException());

                }else {
                    JFDialogs.create()
                            .owner(ComponentManagerPropertiesPane.this)
                            .message(String.format("An error occurred while fetching the advertisement RSpec from %s.",
                                    componentManagerInfo.getAuthority().getName()))
                            .masthead("Fetching advertisement RSpec failed")
                            .title("Fetching advertisement RSpec failed")
                            .showException(updateAdvTask.getException());

                }
            }
        });
    }

    public boolean validate() {
        if (componentIdNodeRadioButton.isSelected() && componentIdComboBox.getSelectionModel().isEmpty()) {
            JFDialogs.create()
                    .owner(this)
                    .message("Please select a node from the list")
                    .masthead("No node selected")
                    .showError();
            return false;
        }

        return true;
    }

    public FXRspecNode getSelectedNode() {
        if (!componentIdNodeRadioButton.isSelected())
            return null;

        return componentIdComboBox.getSelectionModel().getSelectedItem();
    }

    public ReadOnlyBooleanProperty isSpecificNodeChosen() {
        return componentIdNodeRadioButton.selectedProperty();
    }

    public HardwareType getSelectedHardwareType() {
        if (!hardwareTypeNodeRadioButton.isSelected())
            return null;

        return hardwareTypesComboBox.getSelectionModel().getSelectedItem();
    }

    @FXML
    private void onRefreshAdvertisementAction() {
        refreshAdvertisement();
    }

    @FXML
    private void showTestbedMap() {
    }

    public boolean isActiveNodesOnly() {
        return onlyAvailableNodes.isSelected();
    }
}
