package be.iminds.ilabt.jfed.lowlevel.api_wrapper.impl;

import be.iminds.ilabt.jfed.log.Logger;
import be.iminds.ilabt.jfed.lowlevel.*;
import be.iminds.ilabt.jfed.lowlevel.api.AbstractFederationApi;
import be.iminds.ilabt.jfed.lowlevel.api.AbstractFederationApi2;
import be.iminds.ilabt.jfed.lowlevel.api.FederationMemberAuthorityApi2;
import be.iminds.ilabt.jfed.lowlevel.api.FederationSliceAuthorityApi2;
import be.iminds.ilabt.jfed.lowlevel.api_wrapper.UserAndSliceApiWrapper;
import be.iminds.ilabt.jfed.lowlevel.authority.SfaAuthority;
import be.iminds.ilabt.jfed.lowlevel.connection.JFedConnectionProvider;
import be.iminds.ilabt.jfed.lowlevel.connection.SfaConnection;
import be.iminds.ilabt.jfed.preferences.JFedPreferences;
import be.iminds.ilabt.jfed.util.*;
import org.slf4j.LoggerFactory;

import javax.annotation.Nullable;
import java.security.cert.X509Certificate;
import java.text.ParseException;
import java.util.*;

/**
 * ProtoGeniSAUserAndSliceApiWrapper
 */
public class UniformFederationApi2UserAndSliceApiWrapper extends UserAndSliceApiWrapper {
    private static final org.slf4j.Logger LOG =
            LoggerFactory.getLogger(UniformFederationApi2UserAndSliceApiWrapper.class);
    List<AnyCredential> userCredentials;

    public UniformFederationApi2UserAndSliceApiWrapper(Logger logger, GeniUserProvider geniUserProvider, JFedConnectionProvider connectionProvider, JFedPreferences jFedPreferences) {
        super(logger, geniUserProvider, connectionProvider, jFedPreferences);
    }

    private FederationMemberAuthorityApi2 ma(Logger logger) {
        return new FederationMemberAuthorityApi2(logger, jFedPreferences);
    }

    private FederationSliceAuthorityApi2 sa(Logger logger) {
        return new FederationSliceAuthorityApi2(logger, jFedPreferences);
    }

    private SfaConnection getSaConnection() throws JFedException {
        return getConnection(new ServerType(ServerType.GeniServerRole.GENI_CH_SA, 2));
    }

    private SfaConnection getMaConnection() throws JFedException {
        return getConnection(new ServerType(ServerType.GeniServerRole.GENI_CH_MA, 2));
    }

    public FederationSliceAuthorityApi2.GetVersionSAResult getGetVersionSAResult(Logger logger) throws JFedException {
        //TODO: buffer

        AbstractFederationApi2.FederationApiReply<FederationSliceAuthorityApi2.GetVersionSAResult> reply =
                sa(logger).getVersion(getSaConnection());

        if (!reply.getGeniResponseCode().isSuccess())
            throw new JFedException("GetVersion call to SA not successful: code=" +
                    reply.getGeniResponseCode() + " output=" + reply.getOutput(),
                    reply
            );

        return reply.getValue();
    }

    public List<AnyCredential> makeCredentialList(AnyCredential c) {
        List<AnyCredential> res = new ArrayList<AnyCredential>();
        if (c != null)
            res.add(c);
        return res;
    }

    @Override
    public synchronized List<AnyCredential> getUserCredentials(Logger logger, GeniUrn user) throws JFedException {
        userCredentials = null;

        List<AnyCredential> reqCredCreds = new ArrayList<AnyCredential>();

        if (speaksForCredentials != null && !speaksForCredentials.isEmpty()) {
            reqCredCreds.addAll(speaksForCredentials);
            assert speakingForUrn != null;
            user = speakingForUrn;
        }

        assert user != null;

        AbstractFederationApi2.FederationApiReply<List<AnyCredential>> reply =
                ma(logger).getCredentials(
                        getMaConnection(),
                        reqCredCreds,
                        user,
                        addCredentialExtraOptions(null)/*extraOptions*/);

        assert reply != null;

        if (!reply.getGeniResponseCode().isSuccess())
            throw new JFedException("Could not retrieve user credential from server: code=" +
                    reply.getGeniResponseCode() + " output=" + reply.getOutput(),
                    reply
            );

        assert reply.getValue() != null;
        assert !reply.getValue().isEmpty();

        userCredentials = new ArrayList<AnyCredential>();

        if (speaksForCredentials != null && !speaksForCredentials.isEmpty())
            userCredentials.addAll(speaksForCredentials);

        userCredentials.addAll(reply.getValue());
        return new ArrayList<AnyCredential>(userCredentials);
    }

    @Override
    public boolean hasUserCredentials() {
        return hasUserCredentials(userCredentials);
    }

    @Override
    public boolean hasSpeaksForSupport() {
        return true;
    }

    @Override
    public List<AnyCredential> getCachedUserCredentialsForAM() {
        if (userCredentials == null)
            return Collections.emptyList();
        else
            return new ArrayList<AnyCredential>(userCredentials);
    }

    @Override
    public List<AnyCredential> getSliceCredentials(Logger logger, GeniUrn sliceUrn) throws JFedException {
        assert userCredentials != null && !userCredentials.isEmpty();
        assert sliceUrn != null;

        AbstractFederationApi2.FederationApiReply<List<AnyCredential>> reply =
                sa(logger).getSliceCredentials(getSaConnection(), userCredentials, sliceUrn, addCredentialExtraOptions(null)/*extraOptions*/);

        if (!reply.getGeniResponseCode().isSuccess())
            throw new JFedException("Could not retrieve slice credential from server: code=" +
                    reply.getGeniResponseCode() + " output=" + reply.getOutput(),
                    reply
            );

        List<AnyCredential> sliceCredentials = new ArrayList<AnyCredential>();
        if (speaksForCredentials != null && !speaksForCredentials.isEmpty())
            sliceCredentials.addAll(speaksForCredentials);
        sliceCredentials.addAll(reply.getValue());
        return sliceCredentials;
    }

    @Override
    public List<GeniUrn> getSlicesForUser(Logger logger, GeniUrn user) throws JFedException {
        List<GeniUrn> res = new ArrayList<GeniUrn>();
        List<String> resStrings = new ArrayList<String>();

        boolean serverSupportsLookupForMemberMatch =
                getUserAuthority().hasSpecialCase(SfaAuthority.SpecialCase.HAS_FEATURE_LOOKUP_SLICE_FOR_MEMBER_MATCH_EXPIRED);

        //match is officially not supported in the API for this call, but it works on the geni portal and is convenient.
        Hashtable extraOptions = null;
        if (serverSupportsLookupForMemberMatch) {
            Hashtable<String, Boolean> match = new Hashtable<String, Boolean>();
            match.put("SLICE_EXPIRED", false);
            extraOptions = new Hashtable();
            extraOptions.put("match", match);
        }

        AbstractFederationApi2.FederationApiReply<List<FederationSliceAuthorityApi2.UrnRoleTuple>> reply =
                sa(logger).lookupForMember(getSaConnection(), "SLICE", user, userCredentials, addCredentialExtraOptions(extraOptions));
        if (reply == null || reply.getGeniResponseCode() == null || !reply.getGeniResponseCode().isSuccess() || reply.getValue() == null)
            return res;
        List<FederationSliceAuthorityApi2.UrnRoleTuple> urnRoleTuples = reply.getValue();
        boolean sawExpiredSlice = false;
        for (FederationSliceAuthorityApi2.UrnRoleTuple urnRoleTuple : urnRoleTuples) {
            if (!urnRoleTuple.isExpired()) {
                res.add(urnRoleTuple.getUrn());
                resStrings.add(urnRoleTuple.getUrn().toString());
            } else
                sawExpiredSlice = true;
        }

        //since it takes only 1 extra call, we'll slo get all other info about the slices (such as expire date)
        Map<String, Object> match2 = new HashMap<String, Object>();
        match2.put("SLICE_URN", new Vector(resStrings));
        if (serverSupportsLookupForMemberMatch || sawExpiredSlice) {

            if (getUserAuthority().hasSpecialCase(SfaAuthority.SpecialCase.BUG_WORKAROUND_LOOKUP_SLICE_EXPIRED_MATCH_BOOLEAN)) {
                match2.put("SLICE_EXPIRED", "false");
            } else
                match2.put("SLICE_EXPIRED", false);
        }

        boolean hasSliverActiveInfo = false;
        AbstractFederationApi.FederationApiReply<AbstractFederationApi.LookupResult> sliceInfoLookupRes =
                sa(logger).lookup(getSaConnection(), "SLICE", userCredentials, match2, null/*filter*/, addCredentialExtraOptions(null)/*extraOptions*/);
        if (sliceInfoLookupRes != null && sliceInfoLookupRes.getValue() != null) {
            List<String> keys = sliceInfoLookupRes.getValue().getKeys();
            for (String key : keys) {
                if (sliceInfoLookupRes.getValue().get(key).containsKey("_EMULAB_SLICE_HAS_SLIVERS")) {
                    hasSliverActiveInfo = true;
                    break;
                }
            }
        }

        if (!hasSliverActiveInfo) {
            try {
                Map<String, Object> match3 = new HashMap<String, Object>();
                match3.put("SLIVER_INFO_SLICE_URN", new Vector(resStrings));
                AbstractFederationApi.FederationApiReply<AbstractFederationApi.LookupResult> sliverInfoLookupRes =
                        sa(logger).lookup(getSaConnection(), "SLIVER_INFO", userCredentials, match3, null/*filter*/, addCredentialExtraOptions(null)/*extraOptions*/);
            } catch (JFedException e) {
                LOG.error("Error fetching SLIVER_INFO. Will be ignored, because it's not crucial.");
            }
        }

        return res;
        //get other info:
//        Map<String, String> match = new HashMap<String, String>();
//        match.put("MEMBER_URN", user.toString());
//        ma.lookupPublicMemberInfo(con, match, null, null);
//        ma.lookupPrivateMemberInfo(con, userCredentials, match, null, null);
    }

    public List<GeniUrn> getUserProjects(Logger logger, GeniUrn user) throws JFedException {
        List<GeniUrn> res = new ArrayList<GeniUrn>();

        AbstractFederationApi2.FederationApiReply<List<FederationSliceAuthorityApi2.UrnRoleTuple>> reply =
                sa(logger).lookupForMember(getSaConnection(),
                        "PROJECT",
                        user,
                        userCredentials,
                        addCredentialExtraOptions(null)/*extraOptions*/);
        if (reply == null || reply.getGeniResponseCode() == null || !reply.getGeniResponseCode().isSuccess() || reply.getValue() == null)
            return res;
        List<FederationSliceAuthorityApi2.UrnRoleTuple> urnRoleTuples = reply.getValue();
        LOG.debug("UniformFederationApi1UserAndSliceApiWrapper#getUserProjects urnRoleTuples.size()=" + urnRoleTuples.size());
        boolean hasExpired = false;
        for (FederationSliceAuthorityApi2.UrnRoleTuple urnRoleTuple : urnRoleTuples) {

            //first a special exception: server might include EXPIRED field in tuple. We can use that info
            if (urnRoleTuple.dictRaw != null && urnRoleTuple.dictRaw.containsKey("EXPIRED")) {
                hasExpired = true;
                if (TextUtil.objectToBoolean(urnRoleTuple.dictRaw.get("EXPIRED"))) {
                    LOG.debug("skipping expired project: " + urnRoleTuple.getUrn());
                    continue;
                }
            }

            res.add(urnRoleTuple.getUrn());
        }

        if (!hasExpired) {
            LOG.debug("doing second call to check for expired projects");

            Map<String, Object> match = new HashMap<String, Object>();
            if (getUserAuthority().hasSpecialCase(SfaAuthority.SpecialCase.HAS_MATCH_PROJECT_EXPIRED))
                match.put("EXPIRED", Boolean.FALSE);
            List<String> filter = new ArrayList<String>();
            filter.add("PROJECT_URN");
            AbstractFederationApi2.FederationApiReply<AbstractFederationApi2.LookupResult> nonExpiredProj = null;
            try {
                nonExpiredProj =
                        sa(logger).lookup(getSaConnection(),
                                "PROJECT",
                                userCredentials,
                                match,
                                filter,
                                addCredentialExtraOptions(null)/*extraOptions*/);
            } catch (AssertionError e) {
                LOG.warn("lookup_projects call failed. Could not check for expired projects. This error will be ignored (possibly leaving expired projects in the list).", e);
                nonExpiredProj = null;
                return res;
            }catch (Exception e) {
                LOG.warn("lookup_projects call failed. Could not check for expired projects. This error will be ignored (possibly leaving expired projects in the list).", e);
                nonExpiredProj = null;
                return res;
            }
            if (nonExpiredProj == null || nonExpiredProj.getGeniResponseCode() == null || !nonExpiredProj.getGeniResponseCode().isSuccess() || nonExpiredProj.getValue() == null) {
                LOG.warn("lookup_projects call failed. Could not check for expired projects. This error will be ignored (possibly leaving expired projects in the list).");
            } else {
                LOG.debug("Found " + nonExpiredProj.getValue().size() + " non expired projects on server.");
                List<GeniUrn> expiredUserProjects = new ArrayList<GeniUrn>();
                for (GeniUrn resUrn : res) {
                    if (nonExpiredProj.getValue().get(resUrn) == null)
                        expiredUserProjects.add(resUrn);
                }
                LOG.debug("found " + expiredUserProjects.size() + " expired projects in " + res.size() + " user projects: will remove them");
                res.removeAll(expiredUserProjects);
            }
        } else {
            LOG.debug("no second call needed to check for expired projects");
        }

        LOG.debug("UniformFederationApi1UserAndSliceApiWrapper#getUserProjects res.size()=" + res.size());
        return res;
    }

    public AbstractFederationApi2.LookupResult getUserIdentifyingInfo(Logger logger, GeniUrn user) throws JFedException {
        Hashtable<String, String> match = new Hashtable<String, String>();
        match.put("MEMBER_URN", getUser().getUserUrnString());

//        Vector<String> filter = new Vector<String>();

        AbstractFederationApi2.FederationApiReply<AbstractFederationApi2.LookupResult> reply =
                ma(logger).lookup(getMaConnection(),
                        "MEMBER",
                        userCredentials,
                        match,
                        null/*filter*/,
                        addCredentialExtraOptions(null)/*extraOptions*/);


        if (reply == null)
            throw new JFedException("Problem looking up user info for " + getUser().getUserUrnString());

        if (reply.getGeniResponseCode() == null || !reply.getGeniResponseCode().isSuccess() || reply.getValue() == null)
            throw new JFedException("Problem looking up user info for " + getUser().getUserUrnString(), reply);

        return reply.getValue();
    }

    @Override
    public List<String> getAggregatesForSlice(Logger logger, List<AnyCredential> sliceCredentials, GeniUrn sliceUrn) throws JFedException {
        AbstractFederationApi.FederationApiReply<AbstractFederationApi.LookupResult> reply =
                sa(logger).lookupSliverInfo(getSaConnection(),
                        sliceCredentials, sliceUrn, null/*match*/, null/*filter*/, addCredentialExtraOptions(null)/*extraOptions*/);

        if (!reply.getGeniResponseCode().isSuccess()) return null;
        AbstractFederationApi.LookupResult lookupRes = reply.getValue();
        if (lookupRes == null) return null;

        List<String> res = new ArrayList<String>();
        List<String> keys = lookupRes.getKeys();
        for (String key : keys) {
            String aggManUrn = (String) lookupRes.get(key).get("SLIVER_INFO_AGGREGATE_URN");
            if (aggManUrn != null)
                res.add(aggManUrn);
        }

        return res;
    }

    @Override
    public Date getSliceExpiration(Logger logger, List<AnyCredential> sliceCredentials, GeniUrn sliceUrn) throws JFedException {
        Map<String, String> match = new HashMap<String, String>();
        match.put("SLICE_URN", sliceUrn.getValue());
        AbstractFederationApi.FederationApiReply<AbstractFederationApi.LookupResult> reply =
                sa(logger).lookupSlice(getSaConnection(),
                        sliceCredentials,
                        match,
                        null/*filter*/,
                        null/*extraOptions*/);

        if (!reply.getGeniResponseCode().isSuccess()) return null;
        AbstractFederationApi.LookupResult lookupRes = reply.getValue();
        if (lookupRes == null) return null;

        List<String> keys = lookupRes.getKeys();
        for (String key : keys) {
            String expirationDateString = (String) lookupRes.get(key).get("SLICE_EXPIRATION");
            if (expirationDateString != null)
                try {
                    return RFC3339Util.rfc3339StringToDate(expirationDateString);
                } catch (ParseException e) {
                    LOG.debug("Failed to parse expiration date. Will ignore. date=\""+expirationDateString+"\"");
                }
        }

        return null;
    }

    @Override
    public boolean isRegisterAggregatesForSliceSupported() {
        return true;
    }

    @Override
    public void registerAggregatesForSlice(
            Logger logger,
            List<AnyCredential> sliceCredentials,
            GeniUrn sliceUrn,
            GeniUrn aggregateUrn,
            Collection<GeniUrn> sliverUrns,
            Date sliverExpirationDate) throws JFedException {
        LOG.debug("UniformFederationApi2UserAndSliceApiWrapper will register: " + sliverUrns.size());
        GeniUrn creatorUrn = getUser().getUserUrn();
        Date expiration = sliverExpirationDate;

        //fall back to slice expiration
        if (expiration == null)
            for (AnyCredential c : sliceCredentials)
                if (c.getExpiresDate() != null && (expiration == null || expiration.after(c.getExpiresDate())))
                    expiration = c.getExpiresDate();

        for (GeniUrn sliverUrn : sliverUrns) {
            LOG.debug("UniformFederationApi2UserAndSliceApiWrapper is registering: " + sliverUrn);
            AbstractFederationApi.FederationApiReply<Hashtable<String, Object>> reply =
                    sa(logger).createSliverInfo(getSaConnection(),
                            sliceCredentials,
                            sliceUrn.toString(),
                            sliverUrn.toString(),
                            aggregateUrn.toString(),
                            creatorUrn.toString(),
                            expiration,
                            null/*fields*/,
                            addCredentialExtraOptions(null)/*extraOptions*/);
        }
    }

    @Override
    public void updateSliverExpirationDateInfoForSlice(
            Logger logger,
            List<AnyCredential> sliceCredentials,
            GeniUrn sliceUrn,
            Collection<GeniUrn> sliverUrns,
            Date newSliverExpirationDate) throws JFedException {
        LOG.debug("UniformFederationApi2UserAndSliceApiWrapper will register: " + sliverUrns.size());
        GeniUrn creatorUrn = getUser().getUserUrn();
        Date expiration = newSliverExpirationDate;

        //fall back to slice expiration
        if (expiration == null)
            for (AnyCredential c : sliceCredentials)
                if (c.getExpiresDate() != null && (expiration == null || expiration.after(c.getExpiresDate())))
                    expiration = c.getExpiresDate();

        Map<String, String> fields = new HashMap<String, String>();
        fields.put("SLIVER_INFO_EXPIRATION", RFC3339Util.dateToRFC3339String(expiration, true/*force zulu*/));

        for (GeniUrn sliverUrn : sliverUrns) {
            LOG.debug("UniformFederationApi2UserAndSliceApiWrapper is registering: " + sliverUrn);
            AbstractFederationApi.FederationApiReply<String> reply =
                    sa(logger).updateSliverInfo(
                            getSaConnection(),
                            sliceCredentials,
                            sliverUrn + "",
                            fields/*fields*/,
                            addCredentialExtraOptions(null)/*extraOptions*/);
        }
    }

    @Override
    public void unregisterAggregatesForSlice(Logger logger, List<AnyCredential> sliceCredentials, GeniUrn sliceUrn, GeniUrn aggregateUrn, Collection<GeniUrn> sliverUrns) throws JFedException {
        GeniUrn creatorUrn = getUser().getUserUrn();
        Date expiration = null;
        for (AnyCredential c : sliceCredentials)
            if (c.getExpiresDate() != null && (expiration == null || expiration.after(c.getExpiresDate())))
                expiration = c.getExpiresDate();

        if (sliverUrns != null) {
            LOG.debug("UniformFederationApi2UserAndSliceApiWrapper will unregister: " + sliverUrns.size());
            //first check which slivers are still registered
            List<String> registeredSliverUrns = null;
            if (sliceUrn != null)
                try {
                    AbstractFederationApi.FederationApiReply<AbstractFederationApi.LookupResult> reply =
                            sa(logger).lookupSliverInfo(getSaConnection(),
                                    sliceCredentials,
                                    sliceUrn,
                                    null /*match*/,
                                    null/*filter*/,
                                    addCredentialExtraOptions(null)/*extraOptions*/);
                    AbstractFederationApi.LookupResult lres = reply.getValue();
                /* example reply:
                {
                "output": "",
                "code": 0,
                "value": {
                              "urn:publicid:IDN+wall2.ilabt.iminds.be+sliver+43869": {
                                     "SLIVER_INFO_AGGREGATE_URN": "urn:publicid:IDN+wall2.ilabt.iminds.be+authority+cm",
                                     "SLIVER_INFO_SLICE_URN": "urn:publicid:IDN+ch.geni.net:fed4firetest3+slice+tstW",
                                     "SLIVER_INFO_EXPIRATION": "2015-01-12T14:10:30Z",
                                     "SLIVER_INFO_CREATOR_URN": "urn:publicid:IDN+ch.geni.net+user+wimvande",
                                     "SLIVER_INFO_CREATION": "",
                                     "SLIVER_INFO_URN": "urn:publicid:IDN+wall2.ilabt.iminds.be+sliver+43869"
                                     }
                              }
                } */
                    if (lres != null) {
                        registeredSliverUrns = new ArrayList<String>(lres.getKeys());
                    }
                } catch (AssertionError t) {
                    LOG.warn("Exception checking registered slivers before deleting them", t);
                /*ignore*/
                    registeredSliverUrns = null;
                } catch (Exception t) {
                    LOG.warn("Exception checking registered slivers before deleting them", t);
                /*ignore*/
                    registeredSliverUrns = null;
                }

            for (GeniUrn sliverUrn : sliverUrns) {
                if (registeredSliverUrns == null || registeredSliverUrns.contains(sliverUrn.toString())) {
                    LOG.debug("UniformFederationApi2UserAndSliceApiWrapper is unregistering: " + sliverUrn);
                    AbstractFederationApi.FederationApiReply<Boolean> reply =
                            sa(logger).deleteSliverInfo(getSaConnection(),
                                    sliceCredentials,
                                    sliverUrn.toString(),
                                    addCredentialExtraOptions(null)/*extraOptions*/);
                } else {
                    LOG.debug("UniformFederationApi2UserAndSliceApiWrapper is skipping delete of SLIVER_INFO because sliver is not registered anymore.");
                }
            }
        } else {
            LOG.debug("UniformFederationApi2UserAndSliceApiWrapper will unregister ALL sliver for "+aggregateUrn);
            List<GeniUrn> registeredAggSliverUrns = null;
            if (sliceUrn != null)
                try {
                    AbstractFederationApi.FederationApiReply<AbstractFederationApi.LookupResult> reply =
                            sa(logger).lookupSliverInfo(getSaConnection(),
                                    sliceCredentials,
                                    sliceUrn,
                                    null /*match*/,
                                    null/*filter*/,
                                    addCredentialExtraOptions(null)/*extraOptions*/);
                    AbstractFederationApi.LookupResult lres = reply.getValue();
                           /* example reply:
                           {
                           "output": "",
                           "code": 0,
                           "value": {
                                         "urn:publicid:IDN+wall2.ilabt.iminds.be+sliver+43869": {
                                                "SLIVER_INFO_AGGREGATE_URN": "urn:publicid:IDN+wall2.ilabt.iminds.be+authority+cm",
                                                "SLIVER_INFO_SLICE_URN": "urn:publicid:IDN+ch.geni.net:fed4firetest3+slice+tstW",
                                                "SLIVER_INFO_EXPIRATION": "2015-01-12T14:10:30Z",
                                                "SLIVER_INFO_CREATOR_URN": "urn:publicid:IDN+ch.geni.net+user+wimvande",
                                                "SLIVER_INFO_CREATION": "",
                                                "SLIVER_INFO_URN": "urn:publicid:IDN+wall2.ilabt.iminds.be+sliver+43869"
                                                }
                                         }
                           } */
                    if (lres != null) {
                        registeredAggSliverUrns = new ArrayList<GeniUrn>();
                        for (Object sliverUrnString : lres.getKeys()) {
                            GeniUrn sliverUrn = GeniUrn.parse(sliverUrnString.toString());
                            if (sliverUrn != null) {
                                Object sliverCmUrn = lres.get(sliverUrn).get("SLIVER_INFO_AGGREGATE_URN");
                                if (sliverCmUrn != null && sliverCmUrn.equals(aggregateUrn.toString()))
                                    registeredAggSliverUrns.add(sliverUrn);
                            }
                        }
                    }
                } catch (AssertionError t) {
                    LOG.warn("Exception searching registered slivers before deleting them", t);
                           /*ignore*/
                    registeredAggSliverUrns = null;
                } catch (Exception t) {
                    LOG.warn("Exception searching registered slivers before deleting them", t);
                           /*ignore*/
                    registeredAggSliverUrns = null;
                }

            if (registeredAggSliverUrns != null)
                for (GeniUrn sliverUrn : registeredAggSliverUrns) {
                    LOG.debug("UniformFederationApi2UserAndSliceApiWrapper is unregistering: " + sliverUrn);
                    AbstractFederationApi.FederationApiReply<Boolean> reply =
                            sa(logger).deleteSliverInfo(getSaConnection(),
                                    sliceCredentials,
                                    sliverUrn.toString(),
                                    addCredentialExtraOptions(null)/*extraOptions*/);
                }
        }
    }

    @Override
    public void shareSlice(Logger logger, GeniUrn sliceUrn, List<AnyCredential> sliceCredentials, GeniUrn userUrn) throws JFedException {
        List<FederationSliceAuthorityApi2.UrnRoleTuple> membersToAdd =
                new ArrayList<FederationSliceAuthorityApi2.UrnRoleTuple>();
        membersToAdd.add(new FederationSliceAuthorityApi2.UrnRoleTuple("SLICE"/*typeName*/, userUrn, "MEMBER"));

        AbstractFederationApi.FederationApiReply<String> reply =
                sa(logger).modifyMembership(getSaConnection(),
                        "SLICE"/*type*/,
                        sliceUrn,
                        membersToAdd,
                        new ArrayList<GeniUrn>() /*membersToRemove*/,
                        new ArrayList<FederationSliceAuthorityApi2.UrnRoleTuple>() /*membersToChange*/,
                        sliceCredentials,
                        addCredentialExtraOptions(null)/*extraOptions*/
                );

        if (reply == null || reply.getGeniResponseCode() == null)
            throw new JFedException("Sharing slice with " + userUrn + " did not succeed: ", reply);

        if (!reply.getGeniResponseCode().isSuccess()) {
            if (reply.getOutput().contains("DUPLICATE")) {
                LOG.warn("Reply to shared slice was \"" + reply.getOutput() + "\". Assuming this means it is successfully shared already.");
                return;
            }
            throw new JFedException("Sharing slice with " + userUrn + " did not succeed: ", reply);
        }
    }

    @Override
    public List<GeniUrn> getUsersForSubAuthority(Logger logger, String projectName) throws JFedException {
        GeniUrn projectUrn =
                GeniUrn.createGeniUrnFromEncodedParts(getUserAuthority().getNameForUrn(), "project", projectName);
        AbstractFederationApi.FederationApiReply<List<FederationSliceAuthorityApi2.UrnRoleTuple>> reply =
                sa(logger).lookupMembers(getSaConnection(), "PROJECT", projectUrn, userCredentials, addCredentialExtraOptions(null)/*extraOptions*/);

        if (reply == null || reply.getGeniResponseCode() == null || !reply.getGeniResponseCode().isSuccess() || reply.getValue() == null)
            if (reply == null)
                throw new JFedException("Error getting users in project " + projectUrn);
            else {
                throw new JFedException("Error getting users in project " + projectUrn, reply);
            }

        List<GeniUrn> res = new ArrayList<GeniUrn>();
        for (FederationSliceAuthorityApi2.UrnRoleTuple tuple : reply.getValue()) {
            res.add(tuple.getUrn());
        }

        return res;
    }

    @Override
    public List<GeniUrn> getUsersForSlice(Logger logger, GeniUrn sliceUrn, List<AnyCredential> sliceCredentials) throws JFedException {
        AbstractFederationApi.FederationApiReply<List<FederationSliceAuthorityApi2.UrnRoleTuple>> reply =
                sa(logger).lookupMembers(getSaConnection(), "SLICE", sliceUrn, sliceCredentials, addCredentialExtraOptions(null)/*extraOptions*/);

        if (reply == null || reply.getGeniResponseCode() == null || !reply.getGeniResponseCode().isSuccess() || reply.getValue() == null)
            if (reply == null)
                throw new JFedException("Error getting users in " + sliceUrn);
            else {
                throw new JFedException("Error getting users in " + sliceUrn, reply);
            }

        List<GeniUrn> res = new ArrayList<GeniUrn>();
        for (FederationSliceAuthorityApi2.UrnRoleTuple tuple : reply.getValue()) {
            res.add(tuple.getUrn());
        }

        return res;
    }

    @Override
    public void addUsersToSlice(Logger logger, GeniUrn sliceUrn, List<AnyCredential> sliceCredentials, List<GeniUrn> userUrns, String role) throws JFedException {
        List<FederationSliceAuthorityApi2.UrnRoleTuple> membersToAdd =
                new ArrayList<FederationSliceAuthorityApi2.UrnRoleTuple>();
        for (GeniUrn userUrn : userUrns)
            membersToAdd.add(new FederationSliceAuthorityApi2.UrnRoleTuple("SLICE", userUrn, role));

        AbstractFederationApi.FederationApiReply<String> reply =
                sa(logger).modifyMembership(getSaConnection(),
                        "SLICE",
                        sliceUrn,
                        membersToAdd,
                        null /*membersToRemove*/,
                        null /*membersToChange*/,
                        sliceCredentials,
                        addCredentialExtraOptions(null)/*extraOptions*/);

        if (reply == null || reply.getGeniResponseCode() == null || !reply.getGeniResponseCode().isSuccess())
            if (reply == null)
                throw new JFedException("Error adding user to " + sliceUrn);
            else {
                throw new JFedException("Error adding user to " + sliceUrn, reply);
            }
    }

    @Override
    public SliceInfo createSlice(Logger logger, String sliceName, Date expirationDate, String subAuthName) throws JFedException {
        LOG.debug("createSlice called");
        //2 steps: create slice, and then get its credential(s)

        FederationSliceAuthorityApi2.GetVersionSAResult gv = getGetVersionSAResult(logger);
        assert gv != null;

        Map<String, String> fields = new HashMap<String, String>();
        fields.put("SLICE_DESCRIPTION", "jFed Experimenter GUI slice '" + sliceName + "' for user " + getUser().getUserUrnString());

        if (gv.getServices().contains("PROJECT")) {
            if (subAuthName == null) {
                List<GeniUrn> userProjects = getUserProjects(logger, getUser().getUserUrn());
                if (userProjects.isEmpty())
                    throw new JFedException("Slice creation requires a project, but user is not a member of any project.");

                if (userProjects.size() <= 0) {
                    LOG.error("User " + getUser().getUserUrnString() + " is member of no projects! Will try to create slice without project, but that will most likely fail.");
                    return null;
                } else {
                    GeniUrn chosenProject = null;
                    if (userProjects.size() > 1) {
                        if (ExtraInfoCallback.getFederationApiProjectSelectionCallback() != null) {
                            LOG.debug("User is member of " + userProjects.size() + " projects. Using callback to let user choose one.");
                            chosenProject = ExtraInfoCallback.getFederationApiProjectSelection(userProjects);
                        } else {
                            LOG.warn("User " + getUser().getUserUrnString() + " is member of multiple projects: " + userProjects + ". " +
                                    "No callback available to let user choose project: Choosing first project as project for this slice.");
                            chosenProject = userProjects.get(0);
                        }
                    } else
                        chosenProject = userProjects.get(0);
                    if (chosenProject != null)
                        fields.put("SLICE_PROJECT_URN", chosenProject.getValue());
                }
            } else {
                GeniUrn project =
                        GeniUrn.createGeniUrnFromEncodedParts(getUser().getUserUrn().getEncodedTopLevelAuthority(), "project", subAuthName);
                fields.put("SLICE_PROJECT_URN", project.getValue());
            }
        }

        if (gv.getFieldsForObject("PROJECT").containsKey("_GENI_SLICE_EMAIL")) {
            AbstractFederationApi2.LookupResult userIdentifyingInfo =
                    getUserIdentifyingInfo(logger, getUser().getUserUrn());

            if (userIdentifyingInfo != null && userIdentifyingInfo.get(getUser().getUserUrn()).containsKey("MEMBER_EMAIL")) {
                String userEmail = (String) userIdentifyingInfo.get(getUser().getUserUrn()).get("MEMBER_EMAIL");
                fields.put("_GENI_SLICE_EMAIL", userEmail);
            } else {
                throw new JFedException("Slice creation requires a user email, but the user email is not known.");
            }
        }

//      TODO  addRequiredSupplementaryFieldsToCreate(fields, "SLICE");

        if (expirationDate != null) {
            fields.put("SLICE_EXPIRATION", RFC3339Util.dateToRFC3339String(expirationDate, true, true, true));
        }

        LOG.debug("createSlice is calling server with sliceName=" + sliceName + " and fields: " + fields);

        AbstractFederationApi2.FederationApiReply<Hashtable<String, Object>> reply =
                sa(logger).createSlice(
                        getSaConnection(),
                        userCredentials,
                        sliceName/* automatically added to "SLICE_NAME" field*/,
                        fields,
                        addCredentialExtraOptions(null)/*extraOptions*/);

        LOG.debug("createSlice got answer from server");

        if (reply != null && reply.getGeniResponseCode() != null && reply.getGeniResponseCode().isSuccess() && reply.getValue() != null) {
            Hashtable<String, Object> sliceInfo = reply.getValue();

//            if (sliceInfo.containsKey("SLICE_CREDENTIAL"?)) {
            //        //TODO slice credential is normally in returned fields? Otherwise, we can retrieve it with a call.
//            }

            if (sliceInfo.containsKey("SLICE_URN")) {
                String sliceUrnStr = sliceInfo.get("SLICE_URN").toString();
                GeniUrn sliceUrn = GeniUrn.parse(sliceUrnStr);
                if (sliceUrn != null) {
                    AbstractFederationApi2.FederationApiReply<List<AnyCredential>> credReply =
                            sa(logger).getSliceCredentials(getSaConnection(), userCredentials, sliceUrn, addCredentialExtraOptions(null)/*extraOptions*/);

                    if (!credReply.getGeniResponseCode().isSuccess())
                        throw new JFedException("Failed to get slice credentials", credReply);


                    List<AnyCredential> sliceCredentials = new ArrayList<AnyCredential>();
                    if (speaksForCredentials != null && !speaksForCredentials.isEmpty())
                        sliceCredentials.addAll(speaksForCredentials);
                    sliceCredentials.addAll(credReply.getValue());

                    return new SliceInfo(sliceName, sliceUrn, sliceCredentials);
                }
                throw new JFedException("Error creating slice, invalid slice urn: " + sliceUrnStr, reply);
            }
            throw new JFedException("Error creating slice, no SLICE_URN returned", reply);
        } else {
            if (reply == null)
                throw new JFedException("Error creating slice");
            else {
                throw new JFedException("Error creating slice", reply);
            }
        }
    }

    @Override
    public List<AnyCredential> renewSlice(Logger logger, List<AnyCredential> sliceCredentials, Date newExpirationDate) throws JFedException {
        GeniUrn sliceUrn = null;
        for (AnyCredential cred : sliceCredentials)
            if (cred instanceof SfaCredential) {
                SfaCredential sfaCredential = (SfaCredential) cred;
                GeniUrn targetUrn = GeniUrn.parse(sfaCredential.getTargetUrn());
                if (targetUrn != null && targetUrn.getEncodedResourceType().equals("slice")) {
                    sliceUrn = targetUrn;
                    break;
                }
            }
        if (sliceUrn == null)
            throw new JFedException("Could not find slice URN in provided slice credential list.");
        Map<String, String> fields = new HashMap<String, String>();
        fields.put("SLICE_EXPIRATION", RFC3339Util.dateToRFC3339String(newExpirationDate, true/*zulu timezone*/, true/*discard subsecond*/, true/*use Z*/));
        AbstractFederationApi.FederationApiReply<String> updateReply = sa(logger).
                updateSlice(getSaConnection(), sliceCredentials, sliceUrn, fields, addCredentialExtraOptions(null)/*extraOptions*/);

        if (!updateReply.getGeniResponseCode().isSuccess())
            throw new JFedException("Failed to update SLICE_EXPIRATION", updateReply);

        AbstractFederationApi2.FederationApiReply<List<AnyCredential>> credReply =
                sa(logger).getSliceCredentials(getSaConnection(), userCredentials, sliceUrn, addCredentialExtraOptions(null)/*extraOptions*/);

        if (!credReply.getGeniResponseCode().isSuccess())
            throw new JFedException("Failed to get updated slice credentials", credReply);

        List<AnyCredential> newSliceCredentials = new ArrayList<AnyCredential>();
        if (speaksForCredentials != null && !speaksForCredentials.isEmpty())
            newSliceCredentials.addAll(speaksForCredentials);
        newSliceCredentials.addAll(credReply.getValue());
        return newSliceCredentials;
    }

    @Override
    public List<String> getSshKeysForUser(Logger logger, GeniUrn userUrn) throws JFedException {
        Map<String, String> match = new HashMap<String, String>();
        match.put("KEY_MEMBER", userUrn.toString());
        AbstractFederationApi.FederationApiReply<AbstractFederationApi.LookupResult> reply =
                ma(logger).lookupKeyForMember(getMaConnection(), userCredentials, userUrn, null/*match*/, null/*filter*/, addCredentialExtraOptions(null)/*extraOptions*/);
        if (!reply.getGeniResponseCode().isSuccess())
            throw new JFedException("Failed to get SSH keys", reply);

        AbstractFederationApi.LookupResult lRes = reply.getValue();
        List<String> res = new ArrayList<String>();
        for (Object o : lRes.getRaw().values()) {
            Hashtable<String, Object> ht = (Hashtable<String, Object>) o;
            Object publicKey = ht.get("KEY_PUBLIC");
            if (publicKey != null)
                res.add(publicKey + "");
        }

        Map<String, String> matchMember = new HashMap<String, String>();
        matchMember.put("MEMBER_URN", userUrn.toString());
        AbstractFederationApi.FederationApiReply<AbstractFederationApi.LookupResult> reply2 =
                ma(logger).lookupMember(getMaConnection(), userCredentials, matchMember/*match*/, null/*filter*/, addCredentialExtraOptions(null)/*extraOptions*/);
        try {
            if (!reply.getGeniResponseCode().isSuccess())
                throw new JFedException("Failed to get SSH keys from user certificate because lookup MEMBER " + userUrn + " call failed",
                        reply);

            String x509pemCerts = null;
            String geniKey = "_GENI_MEMBER_SSL_CERTIFICATE";
            String emulabKey = "_EMULAB_MEMBER_SSL_CERTIFICATE";
            AbstractFederationApi.LookupResult lookupResult = reply2.getValue();
            if (lookupResult == null || lookupResult.size() != 1)
                throw new JFedException("Failed to get SSH keys from user certificate because " +
                        "lookup MEMBER " + userUrn + " call does not contain a member ssl certificate", reply);
            if (lookupResult.get(0).get(geniKey) != null) {
                x509pemCerts = (String) lookupResult.get(0).get(geniKey);
            }
            if (lookupResult.get(0).get(emulabKey) != null) {
                x509pemCerts = (String) lookupResult.get(0).get(emulabKey);
            }
            //add certificate ssh key
            LOG.debug("Looked up member in order to find public key in ssl certificate");
            if (x509pemCerts == null)
                throw new JFedException("Failed to get SSH keys from user certificate because " +
                        "lookup MEMBER " + userUrn + " call does not contain a member ssl certificate. " +
                        "(Searched for keys: " + geniKey + ", " + emulabKey + ")", reply);
            List<X509Certificate> certs = KeyUtil.pemToX509CertificateChain(x509pemCerts);
            for (X509Certificate cert : certs) {
                //a bit of a hack, to find the user certificate in the chain
                List<GeniUrn> subjectUrns =
                        KeyUtil.findUrnsInCertAltNames(cert, KeyUtil.AltNamesSource.SUBJECT_ALT_NAMES, false);
                LOG.debug("subjectUrns in cert: " + subjectUrns);
                for (GeniUrn subjectUrn : subjectUrns) {
                    if (subjectUrn.getResourceType().equals("user")) {
                        String openSshKey = KeyUtil.publicKeyToOpenSshAuthorizedKeysFormat(cert.getPublicKey());
                        res.add(openSshKey);
                        LOG.debug("Got user key from certificate: " + openSshKey);
                        break;
                    }
                }
            }
        } catch (Exception e) {
            LOG.warn("Failed to get SSH keys from user certificate", e);
        }

        //remove duplicate keys
        LOG.debug("Keys (" + res.size() + ") seen for " + userUrn + ": " + res);
        List<String> uniqueRes = new ArrayList<String>();
        Set<String> seenKeys = new HashSet<String>();
        for (String key : res) {
            String[] parts = key.split(" ");
            if (parts.length >= 2 && parts[0].startsWith("ssh-")) {
                boolean added = seenKeys.add(parts[1]);
                if (added)
                    uniqueRes.add(key);
                else
                    LOG.debug("Duplicate key removed: " + key);
            } else {
                LOG.warn("Unexpected key format seen: " + key);
                uniqueRes.add(key);
            }
        }

        return uniqueRes;
    }


    /* sub authority support */
    @Override
    public SubAuthoritySupport getSubAuthoritySupport(Logger logger) throws JFedException {
        FederationSliceAuthorityApi2.GetVersionSAResult getSAVersionRes = getGetVersionSAResult(logger);
        if (!getSAVersionRes.getServices().contains("PROJECT"))
            return SubAuthoritySupport.SUB_AUTHORITY_FORBIDDEN;
        else
            return SubAuthoritySupport.SUB_AUTHORITY_MANDATORY;
    }

    @Override
    public List<String> getSubAuthorityNames(Logger logger, GeniUrn user) throws JFedException {
        List<String> res = new ArrayList<String>();

        List<GeniUrn> projectUrns = getUserProjects(logger, user);
        for (GeniUrn projectUrn : projectUrns) {
            res.add(projectUrn.getEncodedResourceName());
        }

        LOG.debug("returning subauthorities: " + res);

        return res;
    }


    @Override
    public void setSpeaksFor(List<AnyCredential> speaksForCredentials, @Nullable GeniUrn speakingForUrn) {
        super.setSpeaksFor(speaksForCredentials, speakingForUrn);
        if (speakingForUrn != null)
            setExtraOptionsForCallsWithCredential(Collections.singletonMap("speaking_for", speakingForUrn.getValue()));
    }
}
